@@ -20,6 +20,11 @@ class Local(object):
         self._locked = False
         self.load()
 
+        if 'PATH' in os.environ:
+            self.system_path = os.environ['PATH'].strip(os.pathsep).split(os.pathsep)
+        else:
+            self.system_path = os.defpath.strip(os.pathsep).split(os.pathsep)
+
     def __del__(self):
         if self._locked: self.unlock()
 
@@ -66,7 +71,7 @@ class Local(object):
         ordered = self.roles + self.mixed_roles
         ordered.sort(key=lambda x: -self.role(x)._order)
         return ordered
-    
+
     def load(self):
         '''Load the saved host data from the serialized snapshot file'''
         save_path = os.path.join(self.path, '.save')
@@ -80,7 +85,7 @@ class Local(object):
     def save(self):
         '''Save this host's current options to a serialized snapshot file'''
         unlock = self.lock()
-        
+
         to_save = ['repo', 'roles', 'last_install', 'install_locked']
         saved = {}
         for attr in to_save:
@@ -99,7 +104,7 @@ class Local(object):
     def mirror(self, roles=None, done=None, repo_roles=None):
         '''Find the roles given in this host's repository and sync changes down to the host cache'''
         if repo_roles is None: repo_roles = self.repo.all_roles
-        
+
         if roles is None:
             self.roles = [role for role in self.roles if role in repo_roles]
             roles = self.roles
@@ -108,7 +113,7 @@ class Local(object):
 
         if len(roles) == 0:
             return
-        
+
         # Accumulate the finished roles so we don't double-sync or fail on circular mixins
         if done is None:
             done = {}
@@ -118,14 +123,14 @@ class Local(object):
             return
 
         self.repo.copy_down('', self.path, ['/' + role + '/' for role in mirror_roles] + ['/' + role + '/**' for role in mirror_roles])
-        
+
         mixin_roles = []
         for role in mirror_roles:
             done[role] = True
             mixin_roles.extend([mix for mix in self.role(role).mixins if mix not in done and mix not in mixin_roles])
 
         self.mirror(mixin_roles, done, repo_roles)
-    
+
     def role(self, name):
         '''Given a name, make sure we have a Role instance, cached by name per host'''
         if isinstance(name, role.Role):
@@ -135,7 +140,7 @@ class Local(object):
             self._roles[name] = role.Role(self, name)
 
         return self._roles[name]
-    
+
     def mark_role_order(self, mark_roles=None):
         '''Go through all roles and mark the order field based on the role's installation depth'''
         def recurse_order(roles, order=1):
@@ -170,7 +175,7 @@ class Local(object):
             return
         path = util.get_repopath()
         roles = [role for role in os.listdir(path) if not role.startswith('.') and os.path.isdir(os.path.join(path, role))]
-        
+
         snap_path = os.path.join(path, '.bishop.files')
         try:
             f = open(snap_path, 'r')
@@ -181,7 +186,7 @@ class Local(object):
 
         for role in roles:
             changed, deleted = self.sync_changes(os.path.join(path, role), prev_files)
-            
+
             if len(deleted) > 0:
                 #Deletes cannot be rsynced, so just sync the whole changed role
                 self.repo.copy_up(os.path.join(path, role) + '/', role + '/')
@@ -199,7 +204,7 @@ class Local(object):
     def snapshot_wc(self, path):
         '''Take a snapshot of the files and their modification times so we can be more targetted in what we sync back up to the repo - some multi-user protection'''
         roles = [role for role in os.listdir(path) if not role.startswith('.') and os.path.isdir(os.path.join(path, role))]
-        
+
         snapshot = {}
         for role in roles:
             for root, dirs, files in os.walk(os.path.join(path, role)):
@@ -214,15 +219,15 @@ class Local(object):
                 if len(files) == 0 and len(dirs) == 0 and root not in snapshot:
                     #Orphan directories should be snapshotted
                     snapshot[root] = os.stat(root).st_mtime
-        
+
         snap_path = os.path.join(path, '.bishop.files')
         f = open(snap_path, 'w')
         pickle.dump(snapshot, f)
         f.close()
-    
+
     def sync_changes(self, path, prev_files):
         '''Compare the current tree with the previous snapshot to get a list of changed files within a given role'''
-        
+
         new_role = True
         found_files = []
         changed = []
@@ -254,7 +259,7 @@ class Local(object):
                     found_files.append(root)
                     if '/' in rel_root and root not in prev_files:
                         changed.append(root + '/')
-                
+
                 if '/' not in rel_root:
                     new_role = False # Empty role directories are not new roles
 
@@ -271,9 +276,10 @@ class Local(object):
         unlock = self.lock()
         if roles is None:
             roles = self.roles
-        
+
         self._owned = {}
         self._ready_packages = []
+        self._ready_package_repos = ['default']
         self._ready_prepackage = {}
         shutil.rmtree(self.build_path, True)
 
@@ -286,7 +292,7 @@ class Local(object):
             os.makedirs(os.path.join(meta_path, '.prepackage'))
         except OSError:
             pass
-        
+
         for name, script_path in self._ready_prepackage.items():
             os.symlink(os.path.abspath(script_path), os.path.join(meta_path, '.prepackage', name))
 
@@ -302,6 +308,13 @@ class Local(object):
         if not name in self._ready_packages:
             self._ready_packages.append(name)
 
+        if ':' in name:
+            repo, _ = name.split(':', 1)
+            if repo not in package.repos:
+                raise Exception('Unknown package type: %s - not one of %s' % (repo, list(package.repos.keys())))
+            if repo not in self._ready_package_repos:
+                self._ready_package_repos.append(repo)
+
     def ready_prepackage(self, path, name):
         '''Add a prepackage script to be run before a given package is installed'''
         self._ready_prepackage[name] = path
@@ -316,7 +329,28 @@ class Local(object):
             roles = self.roles
         elif save:
             self.roles = list(set(self.roles).union(set(roles)))
-        
+
+        roles = self.install_preflight(roles, local, cleanup=False)
+
+        change_files = self.get_changes(path)
+        self.run_preinstalls(change_files)
+        [self.install_file(file, path) for file in change_files]
+        self.run_postinstalls(change_files)
+
+        self.last_install = time.time()
+        self._install_roles = None
+        self.save()
+        if unlock: self.unlock()
+
+    def install_preflight(self, roles=None, local=False, cleanup=True):
+        '''Build the given roles, then run a preflight install of just the packages and prepackage scripts on the host'''
+        if self.install_locked and not local:
+            raise Exception('install cancelled - local host is locked for testing')
+
+        unlock = self.lock()
+        if roles is None:
+            roles = self.roles
+
         self._install_roles = roles
         # Roles should be added general-to-specific, but installed specific-to-general
         roles.reverse()
@@ -328,16 +362,13 @@ class Local(object):
         change_packages = self.get_new_packages()
         self.run_prepackages(change_packages)
         self.install_packages(change_packages)
-        
-        change_files = self.get_changes(path)
-        self.run_preinstalls(change_files)
-        [self.install_file(file, path) for file in change_files]
-        self.run_postinstalls(change_files)
-        
-        self.last_install = time.time()
-        self._install_roles = None
-        self.save()
+
+        if cleanup:
+            self._install_roles = None
+            self.save()
+
         if unlock: self.unlock()
+        return roles
 
     def uninstall(self, roles):
         '''Remove a set of roles from the roles this host checks'''
@@ -348,10 +379,22 @@ class Local(object):
 
     def get_new_packages(self):
         '''Return a list of only those packages that have not already been installed'''
-        packager = package.Package()
-        already_installed = packager.list()
-        return [name for name in self._ready_packages if name not in already_installed]
-        
+        already_installed = {}
+        for repo in self._ready_package_repos:
+            already_installed[repo] = package.repos[repo]().list()
+
+        new_packages = []
+        for name in self._ready_packages:
+            if ':' in name:
+                repo, repo_name = name.split(':', 1)
+            else:
+                repo = 'default'
+                repo_name = name
+            if repo_name not in already_installed[repo]:
+                new_packages.append(name)
+
+        return new_packages
+
     def run_prepackages(self, deps):
         '''Gather a unique list of the prepackage script to run based on what packages still need to be installed, and run them'''
         prepackages = {}
@@ -372,10 +415,19 @@ class Local(object):
             util.shell(shell_args)
 
     def install_packages(self, deps):
-        packager = package.Package()
+        packagers = {}
+        for repo in self._ready_package_repos:
+            packagers[repo] = package.repos[repo]()
+
         for dep in deps:
+            if ':' in dep:
+                repo, repo_name = dep.split(':', 1)
+            else:
+                repo = 'default'
+                repo_name = dep
+
             logging.info('package install %s' % dep)
-            packager.install(dep)
+            packagers[repo].install(repo_name)
 
     def install_file(self, file_path, root):
         '''Atomically write the changed file to the given installation path'''
@@ -401,9 +453,20 @@ class Local(object):
         if not os.path.exists(parent_dir):
             util.chown_makedirs(parent_dir)
 
+        install_exec = False
+        for syspath in self.system_path:
+            if parent_dir.rstrip('/') == syspath.rstrip('/'):
+                install_exec = True
+                break
+
+        if install_exec:
+            install_exec_label = 'x'
+        else:
+            install_exec_label = ''
+
         # Pull the uid and gid from the parent directory
         st = os.stat(os.path.dirname(install_path))
-        logging.info('install file %s (%s:%s)' % (install_path, st.st_uid, st.st_gid))
+        logging.info('install file %s (%s:%s%s)' % (install_path, st.st_uid, st.st_gid, install_exec_label))
 
         if orig_path != build_path:
             # If the built file is a symlink, then copy the file so we can atomically move it to the install location without harming any non-build locations
@@ -411,6 +474,8 @@ class Local(object):
             shutil.copyfile(orig_path, build_path)
         shutil.move(build_path, install_path)
         os.chown(install_path, st.st_uid, st.st_gid)
+        if install_exec:
+            os.chmod(install_path, 0755)
 
     def run_preinstalls(self, changes):
         '''Scan the changed files looking for preinstall scripts and run them.  A preinstall script failure should terminate further installation.'''
@@ -466,7 +531,7 @@ class Local(object):
             ordered.sort(key=lambda x: -self.role(os.path.basename(os.path.dirname(x[0])))._order)
 
         return ordered
-    
+
     def get_changes(self, install_root='/'):
         '''Scan the build directories for files that are different than what's in the install root, using sha1 hashes to compare file contents'''
         changed = []
@@ -490,7 +555,7 @@ class Local(object):
         for root, dirs, files in files_to_walk:
             if '.bishop' in dirs:
                 dirs.remove('.bishop')
-            
+
             rel_root = root.replace(self.build_path, '')
             if rel_root.startswith('/'):
                 rel_root = rel_root[1:]
@@ -513,11 +578,11 @@ class Local(object):
         '''Extract the config files a given uninstalled package into a given role'''
         role_path = os.path.join(util.get_repopath(), role_name)
         package.Package().extract(package_name, os.path.join(role_path, 'files'))
-        
+
         # Only keep the files in /etc (a simple way to isolate configuration from other stuff)
         for bad_path in [os.path.join(role_path, 'files', path) for path in os.listdir(os.path.join(role_path, 'files')) if path != 'etc']:
             shutil.rmtree(bad_path)
-        
+
         # Write the package to the packages list (presume the dependencies will always come along for the ride)
         f = open(os.path.join(role_path, 'packages'), 'w')
         f.write(package_name + '\n')
@@ -552,7 +617,7 @@ class Local(object):
                     partials[name].append(file_path)
                 else:
                     partials[name] = [file_path]
-            
+
             if len(partials) > 0:
                 for name, paths in partials.items():
                     paths.sort()
@@ -607,7 +672,7 @@ class Role(object):
             for host, server in self._hosts.items():
                 ret[host] = getattr(server, attr)(*args)
             return ret
-        
+
         try:
             if hasattr(getattr(Local, attr), '__func__'):
                 return _getter
@@ -615,5 +680,5 @@ class Role(object):
                 return _getter()
         except AttributeError:
             return _getter()
-        
+
 
